#!/usr/bin/env python
"""
`htrc.volumes`

Contains the configuration parser object.
"""
from future import standard_library
standard_library.install_aliases()
from typing import Optional
from configparser import RawConfigParser as ConfigParser, NoSectionError
from codecs import open
import logging
import os.path
import shutil
import time

DEFAULT_PATH = os.path.expanduser('~')
DEFAULT_PATH = os.path.join(DEFAULT_PATH, '.htrc')
if not os.path.exists(DEFAULT_PATH):
    DEFAULT_FILE = os.path.dirname(__file__)
    DEFAULT_FILE = os.path.join(DEFAULT_FILE, '.htrc.default')
    logging.info("Copying default config file to home directory.")
    shutil.copyfile(DEFAULT_FILE, DEFAULT_PATH)


class HtrcDataApiConfig:
    def __init__(self,
                 token: Optional[str] = None,
                 host: Optional[str] = None,
                 port: Optional[int] = None,
                 epr: Optional[str] = None,
                 cert: Optional[str] = None,
                 key: Optional[str] = None) -> None:
        super().__init__()

        self.token = token or get_jwt_token()
        self.host = host or get_dataapi_host()
        self.port = port or get_dataapi_port()
        self.epr = epr or get_dataapi_epr()
        self.cert = cert or get_dataapi_cert()
        self.key = key or get_dataapi_key()


def _get_value(section, key, path=None):
    if path is None:
        path = DEFAULT_PATH

    config = ConfigParser(allow_no_value=True)
    with open(path, encoding='utf8') as configfile:
        config.readfp(configfile)
    try:
        return config.get(section, key)
    except NoSectionError:
        raise EnvironmentError("Config not set for {} {} in {}".format(
            section, key, path))


def get_dataapi_port(path=None):
    port = int(_get_value('data', 'port', path))
    return (port)


def get_dataapi_host(path=None):
    host = _get_value('data', 'host', path)
    return (host)


def get_dataapi_epr(path=None):
    return _get_value('data', 'url', path)


def get_dataapi_cert(path=None):
    return _get_value('data', 'cert', path)


def get_dataapi_key(path=None):
    return _get_value('data', 'key', path)


def get_dataapi_access(path=None):
    return _get_value('data', 'pd_only', path)


def get_idp_host_port(path=None):
    host = _get_value('idp', 'host', path)
    port = _get_value('idp', 'port', path)

    return (host, port)


def get_idp_path(path=None):
    return _get_value('idp', 'url')


def get_idp_url(path=None):
    host, port = get_idp_host_port(path)
    path = get_idp_path(path)
    if port == 443:
        # On HTTPS Default Path
        return "https://{}{}".format(host, path)
    else:
        return "https://{}:{}{}".format(host, port, path)


# Add jwt credential access methods
def get_jwt_token(path=None):

    import htrc.auth
    token = htrc.auth.get_jwt_token()

    return token

def save_jwt_token(token, path=None):

    """
    Saves JWT token in the config file.
    """
    # Default to ~/.htrc
    if path is None:
        path = DEFAULT_PATH

    # Default to expiration of now - force a new token on next request
    #if expiration is None:
        #expiration = time.time()

    # Open and modify existing config file, if it exists.
    config = ConfigParser(allow_no_value=True)
    if os.path.exists(path):
        config.read(path)
    if not config.has_section('jwt'):
        config.add_section('jwt')

    # set token and expiration
    config.set('jwt', 'token', token)
    #config.set('jwt', 'expiration', expiration)

    with open(path, 'w') as credential_file:
        config.write(credential_file)

    return token


def remove_jwt_token(path=None):
    """
    Removes JWT token from the config file.
    """
    # Default to ~/.htrc
    if path is None:
        path = DEFAULT_PATH

    # Open and modify existing config file, if it exists.
    config = ConfigParser(allow_no_value=True)
    if os.path.exists(path):
        config.read(path)
    if not config.has_section('jwt'):
        config.add_section('jwt')
    # set token and expiration
    config.set('jwt', 'token', " ")
    #config.set('jwt', 'expiration', " ")

    with open(path, 'w') as credential_file:
        config.write(credential_file)


def get_credentials(path=None):
    """
    Retrieves the username and password from a config file for the Data API.
    Raises an EnvironmentError if not specified.
    See also: credential_prompt
    """
    client_id = _get_value('idp', 'client_id', path)
    client_secret = _get_value('idp', 'client_secret', path)

    if not client_id and not client_secret:
        logging.error("Config path: {}".format(path))
        raise EnvironmentError("No client_id and client_secret stored in config file.")

    return (client_id, client_secret)


def populate_parser(parser):
    return parser


if __name__ == '__main__':
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser = populate_parser(parser)
    parser.parse_args()




from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str

import codecs
import json
import logging
import os, os.path
import re
from time import sleep
from urllib.request import urlopen
from urllib.error import HTTPError
from urllib.parse import quote_plus, urlencode

import requests

from htrc.util import split_items

def get_volume_metadata(id, marc=False):
    """
    Retrieve item metadata `from the HathiTrust Bibliographic API`_.

    Params:
    :param id: HTID for the volume to be retrieved
    :param marc: Retrieve MARC-XML within JSON return value.

    .. _from the HathiTrust Bibliographic API: https://www.hathitrust.org/bib_api
    """
    biblio_api = "https://catalog.hathitrust.org/api/volumes"
    
    if marc:
        biblio_api += '/full'
    else:
        biblio_api += '/brief'

    url = biblio_api + '/htid/{}.json'.format(id)

    try:
        reader = codecs.getreader('utf-8')
        data = json.load(reader(urlopen(url)))
        if len(data['records']) == 1:
            for item in data['items']:
                if item['htid'] == id:
                    md = data['records'][item['fromRecord']]
                    md.update(item)
                    return md
            else:
                raise ValueError
    except (ValueError, IndexError, HTTPError):
        raise ValueError("No result found for " + id)


def safe_volume_metadata(id, marc=False, sleep_time=1):
    """
    Retrieve item metadata `from the HathiTrust Bibliographic API`_.
    
    Unlike :method volume_metadata:, this function returns an empty dictionary,
    rather than an error when metadata is missing.

    Params:
    :param id: HTID for the volume to be retrieved
    :param marc: Retrieve MARC-XML within JSON return value.

    _ https://www.hathitrust.org/bib_api
    """
    try:
        metadata = get_volume_metadata(id, marc)
        if sleep_time:
            sleep(sleep_time)
        return metadata
    except ValueError as err:
        logging.error(err)
        return dict()

def get_bulk_metadata(ids, marc=False):
    """
    Retrieve item metadata `from the HathiTrust Bibliographic API`_.

    Params:
    :param ids: HTIDs for the volumes to be retrieved
    :param marc: Retrieve MARC-XML within JSON return value.

    .. _from the HathiTrust Bibliographic API: https://www.hathitrust.org/bib_api
    """
    biblio_api = "https://catalog.hathitrust.org/api/volumes"

    if marc:
        biblio_api += '/full'
    else:
        biblio_api += '/brief'

    query = '|'.join(['htid:' + id for id in ids])
    url = biblio_api + '/json/' + query

    metadata = dict()
    try:
        reader = codecs.getreader('utf-8')
        raw = json.load(reader(urlopen(url)))

        for id, data in raw.items():
            id = id.replace('htid:','')
            if len(data['records']) == 1:
                for item in data['items']:
                    if item['htid'] == id:
                        item_md = data['records'][item['fromRecord']]
                        item_md.update(item)
                        metadata[id] = item_md
            else:
                metadata[id] = dict()
    except HTTPError:
        raise RuntimeError("Could not access HT Bibliography API.")

    return metadata

def safe_bulk_metadata(ids, marc=False, sleep_time=1):
    """
    Retrieve bulk item metadata `from the HathiTrust Bibliographic API`_.

    Unlike :method get_bulk_metadata:, this function returns an
    empty dictionary, rather than an error when metadata is missing.

    Params:
    :param ids: HTIDs for the volumes to be retrieved
    :param marc: Retrieve MARC-XML within JSON return value.

    _ https://www.hathitrust.org/bib_api
    """
    try:
        metadata = get_bulk_metadata(ids, marc)
        if sleep_time:
            sleep(sleep_time)
        return metadata
    except ValueError as err:
        logging.error(err)
        return dict()

def get_metadata(ids, output_file=None):
    """
    Retrieves metadata for a folder of folders, where each subfolder is named
    for a HathiTrust ID. This structure is the default structure extracted from
    a Data API request (:method htrc.volumes.get_volumes:).
    """
    ids = [str.strip(id).replace('+', ':').replace('=', '/') for id in ids] # data cleanup

    metadata = dict()
    for segment in split_items(ids, 50):
        items = safe_bulk_metadata(segment)
        metadata.update(items)

    if output_file:
        with open(output_file, 'w') as outfile:
            json.dump(metadata, outfile)

    return metadata

def record_metadata(id, sleep_time=1):
    """
    Retrieve metadata for a HathiTrust Record.
    """
    regex = re.compile('\W')
    url = "http://catalog.hathitrust.org/api/volumes/brief/recordnumber/{0}.json"

    url = url.format(id)
    r = requests.get(url)
    data = r.json()

    # data = data['items'][id]
    items = []
    if data:
        for item in data['items']:
            enum = regex.sub('', str(item.get('enumcron', '')).lower())
            htid = item.get('htid', '')
            items.append((enum, htid))
    else:
        items = []

    sleep(sleep_time)
    return items



def volume_solr_metadata(id, sleep_time=0.1):
    """
    Retrieve metadata from HTRC Solr API.

    The HTRC Solr instance is used only for certain extracted features 
    unavailable in the main HathiTrust Bibliographic API. If you are a
    recipient of a HTRC Advanced Collaborative Support (ACS) grant,
    then you may have to use this function.

    """
    solr = "http://chinkapin.pti.indiana.edu:9994/solr/meta/select/?q=id:%s" % id
    solr += "&wt=json"  # retrieve JSON results
    if sleep_time:
        sleep(sleep_time)  # JUST TO MAKE SURE WE ARE THROTTLED
    try:
        reader = codecs.getreader('utf-8')
        data = json.load(reader(urlopen(solr)))
        return data['response']['docs'][0]
    except (ValueError, IndexError, HTTPError):
        logging.error("No result found for " + id)
        return dict()





#!/usr/bin/env python
"""
`htrc.volumes`

Contains functions to retrieve volumes from the HTRC Data API. 

The functions in this package will not operate unless they are 
executed from an HTRC Data Capsule in Secure Mode. The module 
`htrc.mock.volumes` contains Patch objects for testing workflows.
"""
from __future__ import print_function
from future import standard_library

standard_library.install_aliases()


#from builtins import input
from htrc.models import HtrcPage

import http.client
from io import BytesIO, TextIOWrapper
import json
import os.path
import progressbar

#import re
import socket
import ssl
#import sys
#from time import sleep
#from urllib.request import urlopen
#from urllib.error import HTTPError
from urllib.parse import urlencode
#import xml.etree.ElementTree as ET
from urllib.parse import urlencode
from zipfile import ZipFile  # used to decompress requested zip archives.
from tqdm import tqdm
from htrc.runningheaders import parse_page_structure
from functools import partial
import pandas as pd
#from htrc.lib.cli import bool_prompt
from htrc.util import split_items
import htrc.config
import multiprocessing

import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())


def get_volumes(data_api_config: htrc.config.HtrcDataApiConfig, volume_ids, concat=False, mets=False, buffer_size=128):
    """
    Returns volumes from the Data API as a raw zip stream.

    Parameters:
    :token: An OAuth2 token for the app.
    :volume_ids: A list of volume_ids
    :concat: If True, return a single file per volume. If False, return a single
    file per page (default).
    :host: Data API host
    :port: Data API port
    """
    if not volume_ids:
        raise ValueError("volume_ids is empty.")

    url = data_api_config.epr + "volumes"

    for id in volume_ids:
        if ("." not in id
                or " " in id):
            print("Invalid volume id " + id + ". Please correct this volume id and try again.")

    data = {'volumeIDs': '|'.join(
        [id.replace('+', ':').replace('=', '/') for id in volume_ids])}


    if concat:
        data['concat'] = 'true'

    if mets:
        data['mets'] = 'true'

    # Authorization
    headers = {"Authorization": "Bearer " + data_api_config.token,
               "Content-type": "application/x-www-form-urlencoded"}

    # Create SSL lookup
    # TODO: Fix SSL cert verification
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    #ctx.verify_mode = ssl.CERT_NONE

    # Retrieve the volumes
    httpsConnection = http.client.HTTPSConnection(
        data_api_config.host,
        data_api_config.port,
        context=ctx,
        key_file=data_api_config.key,
        cert_file=data_api_config.cert)

    httpsConnection.request("POST", url, urlencode(data), headers)

    response = httpsConnection.getresponse()

    if response.status is 200:
        body = True
        data = BytesIO()
        bytes_downloaded = 0
        bar = progressbar.ProgressBar(max_value=progressbar.UnknownLength,
                                      widgets=[progressbar.AnimatedMarker(), '    ',
                                               progressbar.DataSize(),
                                               ' (', progressbar.FileTransferSpeed(), ')'])

        while body:
            body = response.read(buffer_size)
            data.write(body)
            bytes_downloaded += len(body)
            bar.update(bytes_downloaded)

        data = data.getvalue()
    else:
        logging.debug("Unable to get volumes")
        logging.debug("Response Code: {}".format(response.status))
        logging.debug("Response: {}".format(response.reason))
        raise EnvironmentError("Unable to get volumes.")

    if httpsConnection is not None:
        httpsConnection.close()

    return data


def get_pages(data_api_config: htrc.config.HtrcDataApiConfig, page_ids, concat=False, mets=False, buffer_size=128):
    """
    Returns a ZIP file containing specfic pages.

    Parameters:
    :data_api_config: The configuration data of the DataAPI endpoint.
    :volume_ids: A list of volume_ids
    :concat: If True, return a single file per volume. If False, return a single
    file per page (default).
    """
    if not page_ids:
        raise ValueError("page_ids is empty.")

    url = data_api_config.epr + "pages"

    for id in page_ids:
        if ("." not in id
                or " " in id):
            print("Invalid volume id " + id + ". Please correct this volume id and try again.")

    data = {'pageIDs': '|'.join(
        [id.replace('+', ':').replace('=', '/') for id in page_ids])}

    if concat and mets:
        print("Cannot set both concat and mets with pages.")
    elif concat:
        data['concat'] = 'true'
    elif mets:
        data['mets'] = 'true'

    # Authorization
    headers = {"Authorization": "Bearer " + data_api_config.token,
               "Content-type": "application/x-www-form-urlencoded"}

    # Create SSL lookup
    # TODO: Fix SSL cert verification
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    #ctx.verify_mode = ssl.CERT_NONE

    # Retrieve the volumes
    httpsConnection = http.client.HTTPSConnection(
        data_api_config.host,
        data_api_config.port,
        context=ctx,
        key_file=data_api_config.key,
        cert_file=data_api_config.cert
    )

    httpsConnection.request("POST", url, urlencode(data), headers)

    response = httpsConnection.getresponse()

    if response.status is 200:
        body = True
        data = BytesIO()
        bytes_downloaded = 0
        bar = progressbar.ProgressBar(max_value=progressbar.UnknownLength,
                                      widgets=[progressbar.AnimatedMarker(), '    ',
                                               progressbar.DataSize(),
                                               ' (', progressbar.FileTransferSpeed(), ')'])

        while body:
            body = response.read(buffer_size)
            data.write(body)
            bytes_downloaded += len(body)
            bar.update(bytes_downloaded)

        data = data.getvalue()
    else:
        logging.debug("Unable to get pages")
        logging.debug("Response Code: ".format(response.status))
        logging.debug("Response: ".format(response.reason))
        raise EnvironmentError("Unable to get pages.")

    if httpsConnection is not None:
        httpsConnection.close()

    return data


#def get_oauth2_token(username, password):
    # make sure to set the request content-type as application/x-www-form-urlencoded
    #headers = {"Content-type": "application/x-www-form-urlencoded"}
    #data = { "grant_type": "client_credentials",
             #"client_secret": password,
             #"client_id": username }
    #data = urlencode(data)

    # create an SSL context
    #ctx = ssl.create_default_context()
    #ctx.check_hostname = False
    #ctx.verify_mode = ssl.CERT_NONE

    # make sure the request method is POST
    #host, port = htrc.config.get_oauth2_host_port()
    #oauth2port = htrc.config.get_oauth2_port()
    #oauth2EPRurl = htrc.config.get_oauth2_url()
    #httpsConnection = http.client.HTTPSConnection(host, oauth2port, context=ctx)
    #httpsConnection.request("POST", oauth2EPRurl + "?" + data, "", headers)

    #response = httpsConnection.getresponse()

    # if response status is OK
    #if response.status == 200:
        #data = response.read().decode('utf8')

        #jsonData = json.loads(data)
        #logging.info("*** JSON: {}".format(jsonData))

        #token = jsonData["access_token"]
        #logging.info("*** parsed token: {}".format(token))


    #else:
        #logging.debug("Unable to get token")
        #logging.debug("Response Code: {}".format(response.status))
        #logging.debug("Response: {}".format(response.reason))
        #logging.debug(response.read())
        #raise EnvironmentError("Unable to get token.")

    #if httpsConnection is not None:
        #httpsConnection.close()


    #return token



def grep_error(file_name, output_dir, pattern, txt_index):
    na_volume = []
    if output_dir.endswith("/"):
        file_path = output_dir + file_name
    else:
        file_path = output_dir + "/" + file_name

    if os.path.isfile(file_path):
        for line in open(file_path):
            if pattern in line:
                na_volume.append(line.split()[txt_index])

    return na_volume


    return na_volume


def _to_htrc_page(page_file, zip):
    with TextIOWrapper(BytesIO(zip.read(page_file)), encoding='utf-8') as page:
        return HtrcPage([line.rstrip() for line in page.readlines()])


def download_volumes(volume_ids, output_dir, concat=False, mets=False, pages=False,
                     remove_headers_footers=False, hf_window_size=6, hf_min_similarity=0.7, skip_removed_hf=False,
                     parallelism=multiprocessing.cpu_count(), batch_size=250, data_api_config=None):
    if not 0 < parallelism <= multiprocessing.cpu_count():
        raise ValueError("Invalid parallelism level specified")

    remove_hf_fun = partial(
        _remove_headers_footers_and_save,
        concat=concat,
        hf_min_similarity=hf_min_similarity,
        hf_window_size=hf_window_size,
        skip_removed_hf=skip_removed_hf,
        output_dir=output_dir
    )

    volume_ids = list(set(volume_ids))  # ensure unique volume ids
    num_vols = len(volume_ids)

    data_api_config = data_api_config or htrc.config.HtrcDataApiConfig()

    os.makedirs(output_dir, exist_ok=True)

    if any((data_api_config.token, data_api_config.host, data_api_config.port)) is not None:
        logging.info("obtained token: %s\n" % data_api_config.token)

        try:
            errors = []
            rights = []

            with tqdm(total=num_vols) as progress, multiprocessing.Pool(processes=parallelism) as pool:
                for ids in split_items(volume_ids, batch_size):
                    if pages:
                        if concat and mets:
                            raise ValueError("Cannot set both concat and mets with pages.")
                        else:
                            data = get_pages(data_api_config, ids, concat and not remove_headers_footers, mets)
                    else:
                        data = get_volumes(data_api_config, ids, concat and not remove_headers_footers, mets)

                    volumes = []

                    with ZipFile(BytesIO(data)) as vols_zip:
                        zip_list = vols_zip.namelist()
                        if 'ERROR.err' in zip_list:
                            errors.append(vols_zip.read('ERROR.err').decode('utf-8'))
                            zip_list.remove('ERROR.err')
                        if 'volume-rights.txt' in zip_list:
                            rights_data = vols_zip.read('volume-rights.txt').decode('utf-8')
                            zip_list.remove('volume-rights.txt')
                            if not rights:
                                rights.append(rights_data)
                            else:
                                # due to the format in which 'volume-rights.txt' is created, we have to skip
                                # the first 4 lines which make up the header of the file, to extract only the
                                # actual volume rights data for accumulation
                                rights.append(''.join(rights_data.splitlines(keepends=True)[4:]))

                        zip_volume_paths = [zip_vol_path for zip_vol_path in zip_list if zip_vol_path.endswith('/')]
                        num_vols_in_zip = len(zip_volume_paths)

                        if not remove_headers_footers:
                            vols_zip.extractall(output_dir, members=zip_list)
                            progress.update(num_vols_in_zip)
                        else:
                            for zip_vol_path in zip_volume_paths:
                                sorted_vol_zip_page_paths = sorted(zip_page_path for zip_page_path in zip_list if zip_page_path.startswith(zip_vol_path) and not zip_page_path.endswith('/'))
                                vol_pages = [_to_htrc_page(page_path, vols_zip) for page_path in sorted_vol_zip_page_paths]
                                volumes.append((zip_vol_path, sorted_vol_zip_page_paths, vol_pages))

                    del data, vols_zip

                    num_missing = batch_size - num_vols_in_zip if num_vols >= batch_size else num_vols - num_vols_in_zip
                    progress.update(num_missing)  # update progress bar state to include the missing volumes also

                    # `volumes` will be empty if `remove_headers_footers=False` since the ZIP was extracted
                    # without further processing
                    if volumes:
                        for _ in pool.imap_unordered(remove_hf_fun, volumes):
                            progress.update()

            na_volumes_all = []

            if errors:
                with open(os.path.join(output_dir, 'ERROR.err'), 'w') as err_file:
                    err_file.write(''.join(errors))

                na_volumes_error = grep_error('ERROR.err', output_dir, 'KeyNotFoundException', -1)
                na_volumes_all.extend(na_volumes_error)

            if rights:
                with open(os.path.join(output_dir, 'volume-rights.txt'), 'w') as rights_file:
                    rights_file.write(''.join(rights))

                if htrc.config.get_dataapi_access() == "true":
                    na_volumes_rights = grep_error('volume-rights.txt', output_dir, ' 3', 0)
                    na_volumes_all.extend(na_volumes_rights)

            num_na = len(na_volumes_all)

            if num_na > 0:
                with open(os.path.join(output_dir, 'volumes_not_available.txt'), 'w') as volumes_na:
                    volumes_na.write("\n".join(str(item) for item in na_volumes_all))

                if num_na < 100:
                    print("\nThe following volume ids are not available. \n Please check volumes_not_available.txt "
                          "for the complete list. ")
                    print('\n'.join(str(item) for item in na_volumes_all))
                else:
                    print("\nThere are {:,} unavailable volumes.\n Please check volumes_not_available.txt "
                          "for the "
                          "complete list. \nTo check the validity of volumes in your workset or volume id file go "
                          "to:\n "
                          "https://analytics.hathitrust.org/validateworkset \n or email us at "
                          "htrc-help@hathitrust.org "
                          "for assistance.".format(num_na))

        except socket.error:
            raise RuntimeError("HTRC Data API time out. Check your inode usage if downloading a large workset. "
                               "Contact HTRC for further help.")

    else:
        raise RuntimeError("Failed to obtain the JWT token.")


def _remove_headers_footers_and_save(vol_data, concat, hf_min_similarity, hf_window_size, skip_removed_hf, output_dir):
    zip_vol_path, sorted_vol_zip_page_paths, vol_pages = vol_data
    clean_volid = zip_vol_path[:-1]

    vol_pages = parse_page_structure(vol_pages, window_size=hf_window_size, min_similarity_ratio=hf_min_similarity)
    pages_body = (page.body for page in vol_pages)
    # save the removed headers/footers for user inspection
    if skip_removed_hf:
        if concat:
            with open(os.path.join(output_dir, clean_volid + '.txt'), 'w', encoding='utf-8') as vol_file:
                vol_file.write('\n'.join(pages_body))
        else:
            vol_path = os.path.join(output_dir, zip_vol_path)
            os.mkdir(vol_path)
            for vol_page_path, page_body in zip(sorted_vol_zip_page_paths, pages_body):
                with open(os.path.join(output_dir, vol_page_path), 'w', encoding='utf-8') as page_file:
                    page_file.write(page_body)
    else:
        if concat:
            with open(os.path.join(output_dir, clean_volid + '.txt'), 'w', encoding='utf-8') as vol_file:
                vol_file.write('\n'.join(pages_body))
        else:
            vol_path = os.path.join(output_dir, zip_vol_path)
            os.mkdir(vol_path)
            for vol_page_path, page_body in zip(sorted_vol_zip_page_paths, pages_body):
                with open(os.path.join(output_dir, vol_page_path), 'w', encoding='utf-8') as page_file:
                    page_file.write(page_body)



        removed_hf = []
        for vol_page_path, vol_page in zip(sorted_vol_zip_page_paths, vol_pages):
            if not (vol_page.has_header or vol_page.has_footer):
                # skip reporting pages that don't have an identified header or footer
                continue
            _, page_name = os.path.split(vol_page_path)
            page_name, _ = os.path.splitext(page_name)
            removed_hf.append({'page': page_name, 'header': vol_page.header, 'footer': vol_page.footer})

        if concat:
            removed_hf_filename = os.path.join(output_dir, clean_volid + '_removed_hf.csv')
        else:
            removed_hf_filename = os.path.join(output_dir, clean_volid, 'removed_hf.csv')

        pd.DataFrame(removed_hf, columns=['page', 'header', 'footer']).to_csv(removed_hf_filename, index=False)


def download(args):
    # extract files
    with open(args.file) as IDfile:
        volumeIDs = [line.strip() for line in IDfile]

    data_api_config = htrc.config.HtrcDataApiConfig(
        token=args.token,
        host=args.datahost,
        port=args.dataport,
        epr=args.dataepr,
        cert=args.datacert,
        key=args.datakey
    )

    return download_volumes(volumeIDs, args.output,
                            remove_headers_footers=args.remove_headers_footers or args.remove_headers_footers_and_concat,
                            concat=args.concat or args.remove_headers_footers_and_concat,
                            mets=args.mets,
                            pages=args.pages,
                            hf_window_size=args.window_size,
                            hf_min_similarity=args.min_similarity_ratio,
                            parallelism=args.parallelism,
                            batch_size=args.batch_size,
                            skip_removed_hf=args.skip_removed_hf,
                            data_api_config=data_api_config)




import sys
if sys.version_info.major == 2:
    from mock import Mock, patch
elif sys.version_info.major == 3:
    from unittest.mock import Mock, patch

import unittest2 as unittest
import json
import os.path

import htrc.workset

class TestWorkset(unittest.TestCase):
    def setUp(self):
        self.test_vols = ['mdp.39015050817181', 'mdp.39015055436151',
            'mdp.39015056169157', 'mdp.39015050161697', 'mdp.39015042791874']

        dirname = os.path.dirname(__file__)
        self.example_file = os.path.join(dirname, 'data/example.jsonld')
        with open(self.example_file, 'r') as infile:
            self.json = json.load(infile)

        self.example_csv = os.path.join(dirname, 'data/example.csv')
        with open(self.example_csv, 'r') as infile:
            self.csv = infile.read().encode('utf-8')

    def test_get_volumes(self):
        """
        Test get_volumes by ensuring that the JSON-LD data structure can be
        extracted.
        """
        vols = htrc.workset.get_volumes(self.json)

        # check that each volume correctly parsed out of JSONLD data
        for vol in self.test_vols:
            self.assertIn(vol, vols)

    def test_load_file(self):
        vols = htrc.workset.load(self.example_file)

        # check that each volume correctly parsed out of JSONLD data
        for vol in self.test_vols:
            self.assertIn(vol, vols)
    
    def test_get_volumes_from_csv(self):
        vols = htrc.workset.get_volumes_from_csv(self.csv)

        # check that each volume correctly parsed out of JSONLD data
        for vol in self.test_vols:
            self.assertIn(vol, vols)

    @patch('htrc.workset.urlopen') 
    def test_load_url_hathitrust(self, urlopen_mock):
        ht_url = "https://babel.hathitrust.org/cgi/mb?a=listis&c=548413090"
        
        # test the default URL with a mock
        response_mock = Mock()
        urlopen_mock.return_value = response_mock
        response_mock.read.return_value = self.csv

        vols = htrc.workset.load_url(ht_url)

        # check that each volume correctly parsed out of CSV data
        for vol in self.test_vols:
            self.assertIn(vol, vols)


    @patch('htrc.workset.urlopen') 
    def test_load_url_htrc(self, urlopen_mock):
        htrc_url = 'https://htrc.hathitrust.org/wsid/123456'

        # test the default URL with a mock
        response_mock = Mock()
        urlopen_mock.return_value = response_mock
        response_mock.read.return_value = json.dumps(self.json).encode('utf-8')

        vols = htrc.workset.load_url(htrc_url)

        # check that each volume correctly parsed out of JSONLD data
        for vol in self.test_vols:
            self.assertIn(vol, vols)

        # Also test a direct URL from the triple store service.
        # Since this is using a mock, it doesn't matter if the service 
        # is up or down
        htrc_url2 = ('http://acbres224.ischool.illinois.edu:8080/' +
            'dcWSfetch/getItems?id=http://htrc.hathitrust.org/wsid/189324102')
        vols = htrc.workset.load_url(htrc_url2)

        # check that each volume correctly parsed out of JSONLD data
        for vol in self.test_vols:
            self.assertIn(vol, vols)
    
    @patch('htrc.workset.urlopen') 
    def test_load_url_error(self, urlopen_mock):
        invalid_url = 'blahblahblah'

        with self.assertRaises(ValueError):
            htrc.workset.load_url(invalid_url)


    @patch('htrc.workset.urlopen') 
    def test_load_hathitrust_collection(self, urlopen_mock):
        ht_url = "https://babel.hathitrust.org/cgi/mb?a=listis&c=548413090"

        # test the default URL with a mock
        response_mock = Mock()
        urlopen_mock.return_value = response_mock
        response_mock.read.return_value = self.csv
        
        vols = htrc.workset.load_hathitrust_collection(ht_url)

        # check that each volume correctly parsed out of CSV data
        for vol in self.test_vols:
            self.assertIn(vol, vols)


        # test three malformed URLs:
        # 1. Missing Collection ID
        ht_url = "https://babel.hathitrust.org/cgi/mb?a=listis"
        with self.assertRaises(ValueError):
            htrc.workset.load_hathitrust_collection(ht_url)
        
        # 2. http vs. https
        ht_url = "http://babel.hathitrust.org/cgi/mb?a=listis&c=548413090"
        with self.assertRaises(ValueError):
            htrc.workset.load_hathitrust_collection(ht_url)

        # 3. Non-Babel URL
        ht_url = "https://htrc.hathitrust.org/cgi/mb?a=listis&c=548413090"
        with self.assertRaises(ValueError):
            htrc.workset.load_hathitrust_collection(ht_url)

    @patch('htrc.workset.load_url') 
    def test_load_with_url(self, load_url_mock):
        ht_url = "https://babel.hathitrust.org/cgi/mb?a=listis&c=548413090"
        load_url_mock.return_value = self.test_vols

        vols = htrc.workset.load(ht_url)
        load_url_mock.assert_called_with(ht_url)


suite = unittest.TestLoader().loadTestsFromTestCase(TestWorkset)
unittest.TextTestRunner(verbosity=2).run(suite)



from builtins import input


def bool_prompt(prompt_str, default=None):
    if default is True:
        default = 'y'
    elif default is False:
        default = 'n'

    result = prompt(prompt_str, options=['y', 'n'], default=default)

    if result == 'y':
        return True
    elif result == 'n':
        return False


def prompt(prompt, options=None, default=None):
    # Construct prompt
    prompt = "\n" + prompt

    if options:
        choices = options[:]
        if default and default in choices:
            default_idx = choices.index(default)
            choices[default_idx] = choices[default_idx].upper()
        prompt += " [{0}]".format('/'.join(choices))
    elif default:
        if isinstance(default,str):
            prompt += " [Default: {0}]".format(default.encode('utf-8'))
        else:
            prompt += " [Default: {0}]".format(default)
    prompt += " "

    # Wait for valid response
    result = None
    while result is None or (options and result not in options):
        result = input(prompt)
        result = result.lower().strip()
        if default and result == '':
            result = default

    return result



#!/usr/bin/env python
"""
`htrc.mock.volumes`

Contains functions to test the volume retrieval from the HTRC Data API.
The download functions will return a sample zip file.

See the core documentation for an example of how to use this library.
"""
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

from builtins import input

from configparser import RawConfigParser as ConfigParser
from io import BytesIO
import os, os.path
from zipfile import ZipFile  # used to decompress requested zip archives.

from htrc.lib.cli import bool_prompt
from htrc.auth import credential_prompt
from htrc.config import save_jwt_token

EXAMPLE_FILE = os.path.join(os.path.dirname(__file__), 'example.zip')

def get_volumes(token, volume_ids, concat=False):
    """
    Returns volumes from the Data API as a raw zip stream.

    Parameters:
    :token: An OAuth2 token for the app.
    :volume_ids: A list of volume_ids
    :concat: If True, return a single file per volume. If False, return a single
    file per page (default).
    """
    if not volume_ids:
        raise ValueError("volume_ids is empty.")

    with open(EXAMPLE_FILE, 'rb') as infile:
        data = infile.read()

    return data

def get_pages(token, page_ids, concat=False):
    """
    Returns a ZIP file containing specfic pages.
    
    Parameters:
    :token: An OAuth2 token for the app.
    :volume_ids: A list of volume_ids
    :concat: If True, return a single file per volume. If False, return a single
    file per page (default).
    """
    if not page_ids:
        raise ValueError("page_ids is empty.")
    
    with open(EXAMPLE_FILE, 'rb') as infile:
        data = infile.read()

    return data

def get_oauth2_token(username, password):
    """
    Returns a sample token for oauth2
    """
    return 'a1b2c3d4e5f6'


def credentials_from_config(path):
    """
    Retrieves the username and password from a config file for the Data API.
    DOES NOT raise an EnvironmentError if path is invalid.
    See also: credential_prompt
    """
    username = None
    password = None

    return (username, password)


def download_volumes(volume_ids, output_dir, username=None, password=None):
    # create output_dir folder, if nonexistant
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    # Retrieve token and download volumes
    token = get_oauth2_token(username, password)
    data = get_volumes(token, volume_ids, False)

    with open(EXAMPLE_FILE, 'rb') as infile:
        myzip = ZipFile(infile)
        myzip.extractall(output_dir)
        myzip.close()


def download(args):
    # extract files
    with open(args.file) as IDfile:
        volumeIDs = [line.strip() for line in IDfile]

    return download_volumes(volumeIDs, args.output, args.username, args.password)




import re
from collections import defaultdict
from typing import List, TypeVar, Set, Iterator, Optional, Tuple, Dict

from htrc.models import Page, PageStructure
from htrc.hf_utils import clean_text, levenshtein, pairwise_combine_within_distance, flatten, group_consecutive_when

T = TypeVar('T', bound=Page)
U = TypeVar('U', bound=PageStructure)


class _Line:
    def __init__(self, text: str, line_number: int, page: Page) -> None:
        self.text = text
        self.line_number = line_number
        self.page = page
        self.cleaned_text = clean_text(text)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, _Line):
            raise NotImplemented

        are_equal = self.page is o.page and self.line_number == o.line_number

        return are_equal

    def __ne__(self, o: object) -> bool:
        return not self == o

    def __hash__(self) -> int:
        line_hash = hash(self.line_number)
        page_hash = hash(self.page)
        hash_value = 31 * line_hash + page_hash

        return hash_value

    def __str__(self) -> str:
        return str((self.line_number, self.cleaned_text))

    def similarity_ratio(self, line: '_Line') -> float:
        ratio = 1 - float(levenshtein(self.cleaned_text, line.cleaned_text)) / max(len(self.cleaned_text),
                                                                                   len(line.cleaned_text))

        return ratio


def parse_page_structure(pages: List[T],
                         window_size: int = 6,
                         min_similarity_ratio: float = 0.7,
                         min_cluster_size: int = 3,
                         max_header_lines: int = 3,
                         max_footer_lines: int = 3) -> List[U]:
    def _get_page_lines(p: T) -> List[_Line]:
        return [_Line(text, line_num, p) for line_num, text in enumerate(p.text_lines)]

    def _cluster_lines(lines: List[Tuple[_Line, _Line]]) -> Set[tuple]:
        cluster_map = {}

        for l1, l2 in lines:
            c1 = cluster_map.get(l1)
            c2 = cluster_map.get(l2)

            if c1 is not None and c2 is not None and c1 is not c2:
                smaller, larger = (c1, c2) if len(c1) < len(c2) else (c2, c1)
                larger.extend(smaller)
                for x in smaller:
                    cluster_map[x] = larger
            elif c1 is not None and c2 is None:
                c1.append(l2)
                cluster_map[l2] = c1
            elif c1 is None and c2 is not None:
                c2.append(l1)
                cluster_map[l1] = c2
            elif c1 is None and c2 is None:
                c = [l1, l2]
                cluster_map[l1] = c
                cluster_map[l2] = c

        return set(map(tuple, cluster_map.values()))

    def _group_lines_by_page(lines: Iterator[_Line]) -> Dict[Page, List[_Line]]:
        lines_grouped_by_page = defaultdict(list)
        for line in lines:
            lines_grouped_by_page[line.page].append(line)

        return lines_grouped_by_page

    def _get_last_header_line(lines: List[_Line]) -> Optional[int]:
        if not lines:
            return None

        return max(l.line_number for l in lines)

    def _get_first_footer_line(lines: List[_Line]) -> Optional[int]:
        if not lines:
            return None

        return min(l.line_number for l in lines)

    def _extract_line_numbers(line: _Line) -> Tuple[_Line, List[int]]:
        numbers = [int(match.group(0)) for match in
                   re.finditer(r"(?:(?<=^)|(?<=\s))\d{1,4}(?=\s|$)", line.text, flags=re.UNICODE)]

        return line, numbers

    def _extract_potential_page_numbers(lines: List[_Line]) -> Tuple[_Line, List[int]]:
        assert len(lines) > 0
        line, numbers = _extract_line_numbers(lines[-1])
        if not numbers and not str.strip(line.text) and len(lines) > 1:
            line, numbers = _extract_line_numbers(lines[-2])

        return line, numbers

    candidate_header_lines = []
    candidate_footer_lines = []

    pages_lines = [_get_page_lines(p) for p in pages]

    for lines in pages_lines:
        # ignore lines that are <4 characters long and/or have no alphabetic characters
        candidate_header_lines.append([l for l in lines[:max_header_lines] if not len(l.cleaned_text) < 4])
        candidate_footer_lines.append([l for l in lines[-max_footer_lines:] if not len(l.cleaned_text) < 4])

    headers_for_comparison = pairwise_combine_within_distance(candidate_header_lines, window_size)
    footers_for_comparison = pairwise_combine_within_distance(candidate_footer_lines, window_size)

    header_line_similarities = []
    for (lines1, lines2) in headers_for_comparison:
        header_line_similarities.extend(
            (l1, l2) for l1 in lines1 for l2 in lines2 if l1.similarity_ratio(l2) >= min_similarity_ratio)

    footer_line_similarities = []
    for (lines1, lines2) in footers_for_comparison:
        footer_line_similarities.extend(
            (l1, l2) for l1 in lines1 for l2 in lines2 if l1.similarity_ratio(l2) >= min_similarity_ratio)

    header_clusters = [cluster for cluster in _cluster_lines(header_line_similarities) if
                       len(cluster) >= min_cluster_size]
    footer_clusters = [cluster for cluster in _cluster_lines(footer_line_similarities) if
                       len(cluster) >= min_cluster_size]

    if not footer_clusters:
        potential_page_numbers = [_extract_potential_page_numbers(lines) for lines in pages_lines if lines]
        potential_page_numbers = [(line, numbers[0]) for line, numbers in potential_page_numbers if len(numbers) == 1]
        potential_clusters = map(lambda group: tuple(map(lambda t: t[0], group)),
                                 group_consecutive_when(potential_page_numbers, lambda x, y: y[1] - x[1] == 1))
        footer_clusters = [cluster for cluster in potential_clusters if len(cluster) >= min_cluster_size]

    header_lines_grouped_by_page = _group_lines_by_page(flatten(header_clusters))
    footer_lines_grouped_by_page = _group_lines_by_page(flatten(footer_clusters))

    last_header_line_pages_map = {p: _get_last_header_line(lines) for p, lines in header_lines_grouped_by_page.items()}
    first_footer_line_pages_map = {p: _get_first_footer_line(lines) for p, lines in
                                   footer_lines_grouped_by_page.items()}

    for page in pages:
        last_header_line = last_header_line_pages_map.get(page)
        first_footer_line = first_footer_line_pages_map.get(page)
        page.__class__ = type('StructuredPage', (page.__class__, PageStructure), {})
        page.num_header_lines = last_header_line + 1 if last_header_line is not None else 0
        page.num_footer_lines = len(page.text_lines) - first_footer_line if first_footer_line is not None else 0

    return pages



#!/usr/bin/env python
from __future__ import print_function

from setuptools import setup, find_packages
from setuptools.command.install import install
import os
import platform
import sys
import atexit
import tarfile


__version__ = '0.1.58b0'

# Installing Cython indepdently bc of numpy version. Pinning packages bc of python3.6
install_requires = ['PyLD', 'future', 'prov', 'unicodecsv', 'progressbar2==3.55.0', 'pandas==1.1.5','requests', 'argparse==1.1', 'Cython','tqdm==4.46.0']
# TODO: migrate to docs confix:, 'sphinx-argparse', 'sphinxcontrib-fulltoc']
if sys.version_info.major == 2:
    install_requires.append('configparser')
    install_requires.append('mock')


def _download_config():
    print("Downloading .htrc file...")

    _config_file_url = 'https://analytics.hathitrust.org/files/.htrc'
    _path = os.path.expanduser('~/.htrc')
    try: 
        from urllib.request import urlretrieve
    except ImportError:
        from urllib import urlretrieve
    urlretrieve(_config_file_url, _path)

    print("\n")


def _install_mallet():
    mallet_path = os.path.expanduser('~/mallet')
    if not os.path.exists(mallet_path):
        print('Installing Mallet ...')
        os.makedirs(mallet_path)
        try: 
            from urllib.request import urlretrieve
        except ImportError:
            from urllib import urlretrieve
        mallet_zip, _ = urlretrieve('http://mallet.cs.umass.edu/dist/mallet-2.0.8RC3.tar.gz')
        mallet_dir = tarfile.open(mallet_zip, "r:gz")
        mallet_dir.extractall(path=mallet_path)
        mallet_dir.close()
        print('\n')


class PostInstallCommand(install, object):
    def __init__(self, *args, **kwargs):
        super(PostInstallCommand, self).__init__(*args, **kwargs)
        atexit.register(_download_config)
        atexit.register(_install_mallet)


setup(
    name='htrc',
    version=__version__,
    description='HathiTrust Research Center API Access',
    author="HathiTrust Research Center",
    author_email="htrc@indiana.edu",
    url='http://analytics.hathitrust.org',
    download_url='http://github.com/htrc/HTRC-PythonSDK',
    keywords=[],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Linguistic",
    ],
    packages=find_packages(),
    install_requires=install_requires,
    include_package_data=True,
    data_files=[('htrc/mock/volumes', ['htrc/mock/volumes/example.zip']),
                ('htrc', ['htrc/.htrc.default'])],
    zip_safe=False,
    entry_points={
        'console_scripts': ['htrc = htrc.__main__:main']
    },
    test_suite="unittest2.collector",
    tests_require=['unittest2'],
    cmdclass={'install': PostInstallCommand}
)



import os
from abc import ABC, abstractmethod
from typing import List


class Page(ABC):
    @property
    @abstractmethod
    def text_lines(self) -> List[str]:
        """
        The lines of text on the page
        """
        pass

    @property
    def text(self) -> str:
        return os.linesep.join(self.text_lines)


class PageStructure(Page, ABC):
    def __init__(self) -> None:
        self.num_header_lines = 0
        self.num_footer_lines = 0

    @property
    def has_header(self) -> bool:
        return self.num_header_lines > 0

    @property
    def has_body(self) -> bool:
        return len(self.text_lines) - self.num_header_lines - self.num_footer_lines > 0

    @property
    def has_footer(self) -> bool:
        return self.num_footer_lines > 0

    @property
    def header_lines(self) -> List[str]:
        return self.text_lines[:self.num_header_lines]

    @property
    def body_lines(self) -> List[str]:
        return self.text_lines[self.num_header_lines:len(self.text_lines) - self.num_footer_lines]

    @property
    def footer_lines(self) -> List[str]:
        return self.text_lines[-self.num_footer_lines:] if self.has_footer else []

    @property
    def header(self) -> str:
        return os.linesep.join(self.header_lines)

    @property
    def body(self) -> str:
        return os.linesep.join(self.body_lines)

    @property
    def footer(self) -> str:
        return os.linesep.join(self.footer_lines)


class HtrcPage(Page):
    def __init__(self, lines: List[str]) -> None:
        self._lines = lines

    @property
    def text_lines(self) -> List[str]:
        return self._lines



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import sys
if sys.version_info.major == 2:
    from mock import Mock, patch, PropertyMock
elif sys.version_info.major == 3:
    from unittest.mock import Mock, patch, PropertyMock

import unittest2 as unittest

import htrc.__main__
import argparse

class TestDownload(unittest.TestCase):
    
    @patch('argparse._sys.argv', ['htrc', 'download', 'mdp.1234567'])
    @patch('htrc.__main__.download')
    def test_raw_volume_id(self, download_mock):
        htrc.__main__.main()
        download_mock.assert_called_once()

    @patch('argparse._sys.argv', ['htrc', 'download', '001423370'])
    @patch('htrc.__main__.download')
    def test_raw_record_id(self, download_mock):
        htrc.__main__.main()
        download_mock.assert_called_once()

    @patch('argparse._sys.argv', ['htrc', 'download', 'https://babel.hathitrust.org/cgi/pt?id=mdp.39015078560078;view=1up;seq=13'])
    @patch('htrc.__main__.download')
    def test_babel_url(self, download_mock):
        htrc.__main__.main()
        download_mock.assert_called_once()
    
    @patch('argparse._sys.argv', ['htrc', 'download', 'https://hdl.handle.net/2027/mdp.39015078560078'])
    @patch('htrc.__main__.download')
    def test_handle_url(self, download_mock):
        htrc.__main__.main()
        download_mock.assert_called_once()

    @patch('argparse._sys.argv', ['htrc', 'download', 'https://catalog.hathitrust.org/Record/001423370'])
    @patch('htrc.__main__.download')
    def test_catalog_url(self, download_mock):
        htrc.__main__.main()
        download_mock.assert_called_once()

    @patch('argparse._sys.argv', ['htrc', 'download', 'https://babel.hathitrust.org/shcgi/mb?a=listis;c=696632727'])
    @patch('htrc.__main__.download')
    def test_collection_builder_url(self, download_mock):
        htrc.__main__.main()
        download_mock.assert_called_once()


import re
from typing import TypeVar, List, Iterator, Tuple, Callable

T = TypeVar('T')


def clean_text(s: str) -> str:
    # replace all characters which aren't letters with whitespaces ([\W\d_] is equivalent of \P{L} which is unsupported)
    s = re.sub(r'[\W\d_]+', " ", s, flags=re.UNICODE)
    # replace multiple sequential whitespaces with single whitespace
    s = re.sub(r'\s{2,}', " ", s, flags=re.UNICODE)
    # trim whitespaces at the beginning and end
    s = s.strip()
    # lowercase
    s = s.lower()

    return s


def levenshtein(s: str, t: str, insert_cost: int = 1, delete_cost: int = 1, replace_cost: int = 1) -> int:
    """ From Wikipedia article; Iterative with two matrix rows. """
    # degenerate cases
    if s == t:
        return 0

    len0 = len(s)
    len1 = len(t)

    if not len0:
        return len1

    if not len1:
        return len0

    # the array of distances
    v0 = [0] * (len0 + 1)
    v1 = [0] * (len0 + 1)

    # initial cost of skipping prefix in s
    for i in range(len(v0)):
        v0[i] = i

    # dynamically compute the array of distances

    # transformation cost for each letter in t
    for j in range(len1):
        # initial cost of skipping prefix in t
        v1[0] = j + 1

        # transformation cost for each letter in s
        for i in range(len0):
            # matching current letters in both strings
            match = 0 if s[i] == t[j] else 1

            # computing cost for each transformation
            cost_insert = v0[i + 1] + insert_cost
            cost_delete = v1[i] + delete_cost
            cost_replace = v0[i] + match * replace_cost

            # keep minimum cost
            v1[i + 1] = min(cost_insert, cost_delete, cost_replace)

        # swap cost arrays
        v0, v1 = v1, v0

    # the distance is the cost for transforming all letters in both strings
    return v0[len0]


def pairwise_combine_within_distance(xs: List[T], n: int) -> List[Tuple[T, T]]:
    if not xs:
        return []

    result = []
    x, xs = xs[0], xs[1:]

    while xs:
        result = result + [(x, v) for v in xs[:n - 1]]
        x, xs = xs[0], xs[1:]

    return result


def group_consecutive_when(xs: List[T], pred: Callable[[T, T], bool]) -> Iterator[List[T]]:
    result = []
    _prev, _next = None, None

    while len(xs) > 1:
        _prev, _next = xs[0], xs[1]
        result.append(_prev)
        if not pred(_prev, _next):
            yield result
            result = []
        xs = xs[1:]

    if len(xs) == 1:
        _prev, _next = _next, xs[0]

    if _prev is not None and _next is not None and pred(_prev, _next):
        result.extend([_prev, _next])
    elif _next is not None:
        result.append(_next)

    yield result


def flatten(xss: List[tuple]) -> Iterator[T]:
    for xs in xss:
        for x in xs:
            yield x



#from base64 import b64encode
from getpass import getpass
#import http.client
#import ssl
#import time
import subprocess

import requests
import requests.auth
#import configparser

import htrc.config


def get_jwt_token():
    # Currently we just store one common jwt token locally at .htrc file for simplicity
    # Expect to add POST method to query unique jwt token with the combo of username and password
    #username, password = credential_prompt()

    #client_id, client_secret = htrc.config.get_credentials()

    #auth = requests.auth.HTTPBasicAuth(client_id, client_secret)
    #data = { "grant_type": "password",
             #"username": username,
             #"password": password,
             #"scope" : "openid"}

    url1 = htrc.config.get_idp_url()
    capsule_id = htrc.config._get_value("jwt", "capsule_id")
    result = subprocess.check_output("hostname -s -I | awk '{print $1}'", shell=True)
    result = result.decode('utf-8')
    result = result[:-1]
    capsule_ip = result.strip()
    url = url1 + "/" + capsule_id + "/" + capsule_ip
    r = requests.get(url)

    data = r.json()
    if 'error' not in data:
        #expiration = int(time.time()) + data['expires_in']
        return data['token']
    elif data['error'] == 'invalid_grant':
        print("Invalid username or password. Please try again.\n")
        return get_jwt_token()
    else:
        raise RuntimeError("JWT token retrieval failed: {}".format(data['error']))


def credential_prompt():
    """
    A prompt for entering HathiTrust Research Center credentials.
    """
    print("Please enter your HathiTrust Research Center credentials.")
    username = input("HTRC Username: ")
    password = getpass("HTRC Password: ")

    if not username or not password:
        print("Invalid username or password. Please try again.\n")
        return credential_prompt()
    else:
        return (username, password)


if __name__ == '__main__':
    token = get_jwt_token()
    htrc.config.save_jwt_token(token)
